@@ -180,9 +180,9 @@ nice_address_get_port (const NiceAddress *addr)
       return ntohs (addr->s.ip4.sin_port);
     case AF_INET6:
       return ntohs (addr->s.ip6.sin6_port);
+    default:
+      g_return_val_if_reached (0);
     }
-
-  g_return_val_if_reached (0);
 }
 
 
@@ -226,20 +226,25 @@ nice_address_set_from_sockaddr (NiceAddress *addr,
 
 NICEAPI_EXPORT void
 nice_address_copy_to_sockaddr (const NiceAddress *addr,
-                               struct sockaddr *sa)
+                               struct sockaddr *_sa)
 {
-  struct sockaddr_in *sin4 = (struct sockaddr_in *)sa;
-  struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+  union {
+    struct sockaddr *addr;
+    struct sockaddr_in *in;
+    struct sockaddr_in6 *in6;
+  } sa;
 
-  g_assert (sa);
+  sa.addr = _sa;
+
+  g_assert (_sa);
 
   switch (addr->s.addr.sa_family)
     {
     case AF_INET:
-      memcpy (sin4, &addr->s.ip4, sizeof (*sin4));
+      memcpy (sa.in, &addr->s.ip4, sizeof (*sa.in));
       break;
     case AF_INET6:
-      memcpy (sin6, &addr->s.ip6, sizeof (*sin6));
+      memcpy (sa.in6, &addr->s.ip6, sizeof (*sa.in6));
       break;
     default:
       g_return_if_reached ();
@@ -278,9 +283,10 @@ nice_address_equal (const NiceAddress *a, const NiceAddress *b)
       return IN6_ARE_ADDR_EQUAL (&a->s.ip6.sin6_addr, &b->s.ip6.sin6_addr)
           && (a->s.ip6.sin6_port == b->s.ip6.sin6_port)
           && (a->s.ip6.sin6_scope_id == b->s.ip6.sin6_scope_id);
-    }
 
-  g_return_val_if_reached (FALSE);
+    default:
+      g_return_val_if_reached (FALSE);
+    }
 }
 
 
@@ -345,9 +351,9 @@ nice_address_is_private (const NiceAddress *a)
       return ipv4_address_is_private (a->s.ip4.sin_addr.s_addr);
     case AF_INET6:
       return ipv6_address_is_private (a->s.ip6.sin6_addr.s6_addr);
+    default:
+      g_return_val_if_reached (FALSE);
     }
-
-  g_return_val_if_reached (FALSE);
 }
 
 
